@@ -242,8 +242,17 @@ static int read_and_queue(struct backend_seatd *backend, int *opcode) {
 			}
 			break;
 		default:
-			if (opcode != NULL &&
-			    connection_pending(&backend->connection) >= header.size) {
+			// If we do not have an opcode pointer, the caller only
+			// expected to see background events so this would be
+			// an error and we might as well stop now. Otherwise,
+			// store the opcode once we have the full message.
+			if (opcode == NULL) {
+				log_errorf("Unexpected response: expected background event, got opcode %d",
+					   header.opcode);
+				set_error(backend);
+				errno = EBADMSG;
+				packets = -1;
+			} else if (connection_pending(&backend->connection) >= header.size) {
 				*opcode = header.opcode;
 			}
 			connection_restore(&backend->connection, sizeof header);
@@ -264,6 +273,7 @@ static int poll_connection(struct backend_seatd *backend, int timeout) {
 	}
 
 	if (fd.revents & (POLLERR | POLLHUP)) {
+		set_error(backend);
 		errno = EPIPE;
 		return -1;
 	}
@@ -272,9 +282,11 @@ static int poll_connection(struct backend_seatd *backend, int timeout) {
 	if (fd.revents & POLLIN) {
 		len = connection_read(&backend->connection);
 		if (len == 0) {
+			set_error(backend);
 			errno = EIO;
 			return -1;
 		} else if (len == -1 && errno != EAGAIN) {
+			set_error(backend);
 			return -1;
 		}
 	}
@@ -308,20 +320,6 @@ static int get_fd(struct libseat *base) {
 	return backend->connection.fd;
 }
 
-static int read_and_execute(struct backend_seatd *backend) {
-	int opcode = 0;
-	int dispatched = read_and_queue(backend, &opcode);
-	if (dispatched == -1) {
-		return -1;
-	} else if (opcode != 0) {
-		// We should only receive background events here, but we got a return value.
-		errno = EINVAL;
-		return -1;
-	}
-	dispatched += execute_events(backend);
-	return dispatched;
-}
-
 static int dispatch(struct libseat *base, int timeout) {
 	struct backend_seatd *backend = backend_seatd_from_libseat_backend(base);
 	if (backend->error) {
@@ -329,10 +327,12 @@ static int dispatch(struct libseat *base, int timeout) {
 		return -1;
 	}
 
-	int predispatch = read_and_execute(backend);
+	int predispatch = read_and_queue(backend, NULL);
 	if (predispatch == -1) {
-		return -1;
+		log_errorf("Could not read and queue events: %s", strerror(errno));
+		goto error;
 	}
+	predispatch += execute_events(backend);
 
 	// We don't want to block if we dispatched something, as the
 	// caller might be waiting for the result. However, we'd also
@@ -348,15 +348,20 @@ static int dispatch(struct libseat *base, int timeout) {
 		return predispatch;
 	} else if (read == -1 && errno != EAGAIN) {
 		log_errorf("Could not read from connection: %s", strerror(errno));
-		return -1;
+		goto error;
 	}
 
-	int postdispatch = read_and_execute(backend);
+	int postdispatch = read_and_queue(backend, NULL);
 	if (postdispatch == -1) {
-		return -1;
+		log_errorf("Could not read and queue events: %s", strerror(errno));
+		goto error;
 	}
+	postdispatch += execute_events(backend);
 
 	return predispatch + postdispatch;
+
+error:
+	return -1;
 }
 
 static struct libseat *_open_seat(const struct libseat_seat_listener *listener, void *data, int fd) {
@@ -393,9 +398,26 @@ static struct libseat *_open_seat(const struct libseat_seat_listener *listener,
 		errno = EBADMSG;
 		goto backend_error;
 	}
+	if (rmsg.seat_name_len > MAX_SEAT_LEN) {
+		log_errorf("Invalid message: seat_name too long (%d)", rmsg.seat_name_len);
+		errno = EBADMSG;
+		goto backend_error;
+	}
 	if (conn_get(backend, backend->seat_name, rmsg.seat_name_len) == -1) {
 		goto backend_error;
 	}
+	// handle old seatd gracefully (seat_name without null byte)
+	if (rmsg.seat_name_len == 0 ||
+	    (rmsg.seat_name_len < MAX_SEAT_LEN && backend->seat_name[rmsg.seat_name_len - 1] != 0)) {
+		backend->seat_name[rmsg.seat_name_len] = 0;
+		rmsg.seat_name_len++;
+	}
+	if (rmsg.seat_name_len == 0 ||
+	    strnlen(backend->seat_name, rmsg.seat_name_len) != (uint16_t)(rmsg.seat_name_len - 1)) {
+		log_error("Invalid message: seat_name not null terminated");
+		errno = EBADMSG;
+		goto backend_error;
+	}
 
 	execute_events(backend);
 	return &backend->base;
@@ -417,28 +439,29 @@ static struct libseat *open_seat(const struct libseat_seat_listener *listener, v
 }
 
 static int close_seat(struct libseat *base) {
+	int res = 0;
 	struct backend_seatd *backend = backend_seatd_from_libseat_backend(base);
+	if (backend->error) {
+		res = -1;
+		goto done;
+	}
 
 	struct proto_header header = {
 		.opcode = CLIENT_CLOSE_SEAT,
 		.size = 0,
 	};
 	if (conn_put(backend, &header, sizeof header) == -1 || read_until_response(backend) == -1) {
-		goto error;
+		res = -1;
 	}
 
 	if (read_header(backend, SERVER_SEAT_CLOSED, 0, false) == SIZE_MAX) {
-		goto error;
+		res = -1;
 	}
 
+done:
 	execute_events(backend);
 	destroy(backend);
-	return 0;
-
-error:
-	execute_events(backend);
-	destroy(backend);
-	return -1;
+	return res;
 }
 
 static const char *seat_name(struct libseat *base) {
@@ -458,6 +481,9 @@ static int send_ping(struct backend_seatd *backend) {
 }
 
 static void check_pending_events(struct backend_seatd *backend) {
+	if (read_and_queue(backend, NULL) == -1) {
+		return;
+	}
 	if (linked_list_empty(&backend->pending_events)) {
 		return;
 	}
@@ -498,21 +524,20 @@ static int open_device(struct libseat *base, const char *path, int *fd) {
 	if (conn_put(backend, &header, sizeof header) == -1 ||
 	    conn_put(backend, &msg, sizeof msg) == -1 || conn_put(backend, path, pathlen) == -1 ||
 	    read_until_response(backend) == -1) {
-		goto error;
+		return -1;
 	}
 
+	int res = 0;
 	struct proto_server_device_opened rmsg;
 	if (read_header(backend, SERVER_DEVICE_OPENED, sizeof rmsg, false) == SIZE_MAX ||
 	    conn_get(backend, &rmsg, sizeof rmsg) == -1 || conn_get_fd(backend, fd)) {
-		goto error;
+		res = -1;
+	} else {
+		res = rmsg.device_id;
 	}
 
 	check_pending_events(backend);
-	return rmsg.device_id;
-
-error:
-	check_pending_events(backend);
-	return -1;
+	return res;
 }
 
 static int close_device(struct libseat *base, int device_id) {
@@ -535,19 +560,16 @@ static int close_device(struct libseat *base, int device_id) {
 	};
 	if (conn_put(backend, &header, sizeof header) == -1 ||
 	    conn_put(backend, &msg, sizeof msg) == -1 || read_until_response(backend) == -1) {
-		goto error;
+		return -1;
 	}
 
+	int res = 0;
 	if (read_header(backend, SERVER_DEVICE_CLOSED, 0, false) == SIZE_MAX) {
-		goto error;
+		res = -1;
 	}
 
 	check_pending_events(backend);
-	return 0;
-
-error:
-	check_pending_events(backend);
-	return -1;
+	return res;
 }
 
 static int switch_session(struct libseat *base, int session) {
@@ -557,6 +579,7 @@ static int switch_session(struct libseat *base, int session) {
 		return -1;
 	}
 	if (session < 0) {
+		errno = EINVAL;
 		return -1;
 	}
 
@@ -569,19 +592,16 @@ static int switch_session(struct libseat *base, int session) {
 	};
 	if (conn_put(backend, &header, sizeof header) == -1 ||
 	    conn_put(backend, &msg, sizeof msg) == -1 || read_until_response(backend) == -1) {
-		goto error;
+		return -1;
 	}
 
+	int res = 0;
 	if (read_header(backend, SERVER_SESSION_SWITCHED, 0, false) == SIZE_MAX) {
-		goto error;
+		res = -1;
 	}
 
 	check_pending_events(backend);
-	return 0;
-
-error:
-	check_pending_events(backend);
-	return -1;
+	return res;
 }
 
 static int disable_seat(struct libseat *base) {
@@ -595,19 +615,16 @@ static int disable_seat(struct libseat *base) {
 		.size = 0,
 	};
 	if (conn_put(backend, &header, sizeof header) == -1 || read_until_response(backend) == -1) {
-		goto error;
+		return -1;
 	}
 
+	int res = 0;
 	if (read_header(backend, SERVER_SEAT_DISABLED, 0, false) == SIZE_MAX) {
-		goto error;
+		res = -1;
 	}
 
 	check_pending_events(backend);
-	return 0;
-
-error:
-	check_pending_events(backend);
-	return -1;
+	return res;
 }
 
 const struct seat_impl seatd_impl = {
